# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
import re
import pytest

from functools import partial
from os import path
from typing import Dict, List, Optional
from urllib.parse import urlparse


logger = logging.getLogger(__name__)


# Check get_local_factory function
# Maximum number of iteration checks to generate requests responses
MAX_VISIT_FILES = 10


def get_response_cb(request, context, datadir,
                    ignore_urls: List[str] = [],
                    visits: Optional[Dict] = None):
    """Mount point callback to fetch on disk the request's content.

    This is meant to be used as 'body' argument of the requests_mock.get()
    method.

    It will look for files on the local filesystem based on the requested URL,
    using the following rules:

    - files are searched in the datadir/<hostname> directory

    - the local file name is the path part of the URL with path hierarchy
      markers (aka '/') replaced by '_'

    Eg. if you use the requests_mock fixture in your test file as:

        requests_mock.get('https://nowhere.com', body=get_response_cb)
        # or even
        requests_mock.get(re.compile('https://'), body=get_response_cb)

    then a call requests.get like:

        requests.get('https://nowhere.com/path/to/resource?a=b&c=d')

    will look the content of the response in:

        datadir/nowhere.com/path_to_resource,a=b,c=d

    Args:
        request (requests.Request): Object requests
        context (requests.Context): Object holding response metadata
            information (status_code, headers, etc...)
        ignore_urls: urls whose status response should be 404 even if the local
            file exists
        visits: Dict of url, number of visits. If None, disable multi visit
            support (default)

    Returns:
        Optional[FileDescriptor] on disk file to read from the test context

    """
    logger.debug('get_response_cb(%s, %s)', request, context)
    logger.debug('url: %s', request.url)
    logger.debug('ignore_urls: %s', ignore_urls)
    if request.url in ignore_urls:
        context.status_code = 404
        return None
    url = urlparse(request.url)
    dirname = url.hostname  # pypi.org | files.pythonhosted.org
    # url.path: pypi/<project>/json -> local file: pypi_<project>_json
    filename = url.path[1:]
    if filename.endswith('/'):
        filename = filename[:-1]
    filename = filename.replace('/', '_')
    if url.query:
        filename += ',' + url.query.replace('&', ',')

    filepath = path.join(datadir, dirname, filename)
    if visits is not None:
        visit = visits.get(url, 0)
        visits[url] = visit + 1
        if visit:
            filepath = filepath + '_visit%s' % visit

    if not path.isfile(filepath):
        logger.debug('not found filepath: %s', filepath)
        context.status_code = 404
        return None
    fd = open(filepath, 'rb')
    context.headers['content-length'] = str(path.getsize(filepath))
    return fd


@pytest.fixture
def datadir(request):
    """By default, returns the test directory's data directory.

    This can be overriden on a per arborescence basis. Add an override
    definition in the local conftest, for example:

        import pytest

        from os import path

        @pytest.fixture
        def datadir():
            return path.join(path.abspath(path.dirname(__file__)), 'resources')


    """
    return path.join(path.dirname(str(request.fspath)), 'data')


def requests_mock_datadir_factory(ignore_urls: List[str] = [],
                                  has_multi_visit: bool = False):
    """This factory generates fixture which allow to look for files on the
    local filesystem based on the requested URL, using the following rules:

    - files are searched in the datadir/<hostname> directory

    - the local file name is the path part of the URL with path hierarchy
      markers (aka '/') replaced by '_'

    Multiple implementations are possible, for example:

    - requests_mock_datadir_factory([]):
        This computes the file name from the query and always returns the same
        result.

    - requests_mock_datadir_factory(has_multi_visit=True):
        This computes the file name from the query and returns the content of
        the filename the first time, the next call returning the content of
        files suffixed with _visit1 and so on and so forth. If the file is not
        found, returns a 404.

    - requests_mock_datadir_factory(ignore_urls=['url1', 'url2']):
        This will ignore any files corresponding to url1 and url2, always
        returning 404.

    Args:
        ignore_urls: List of urls to always returns 404 (whether file
            exists or not)
        has_multi_visit: Activate or not the multiple visits behavior

    """
    @pytest.fixture
    def requests_mock_datadir(requests_mock, datadir):
        if not has_multi_visit:
            cb = partial(get_response_cb,
                         ignore_urls=ignore_urls,
                         datadir=datadir)
            requests_mock.get(re.compile('https://'), body=cb)
        else:
            visits = {}
            requests_mock.get(re.compile('https://'), body=partial(
                get_response_cb, ignore_urls=ignore_urls, visits=visits,
                datadir=datadir)
            )

        return requests_mock

    return requests_mock_datadir


# Default `requests_mock_datadir` implementation
requests_mock_datadir = requests_mock_datadir_factory([])

# Implementation for multiple visits behavior:
# - first time, it checks for a file named `filename`
# - second time, it checks for a file named `filename`_visit1
# etc...
requests_mock_datadir_visits = requests_mock_datadir_factory(
    has_multi_visit=True)



# Copyright (C) 2015  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
import shutil
import pytest
import pkg_resources.extern.packaging.version

from swh.core import config

pytest_v = pkg_resources.get_distribution("pytest").parsed_version
if pytest_v < pkg_resources.extern.packaging.version.parse('3.9'):
    @pytest.fixture
    def tmp_path(request):
        import tempfile
        import pathlib
        with tempfile.TemporaryDirectory() as tmpdir:
            yield pathlib.Path(tmpdir)


default_conf = {
    'a': ('int', 2),
    'b': ('string', 'default-string'),
    'c': ('bool', True),
    'd': ('int', 10),
    'e': ('int', None),
    'f': ('bool', None),
    'g': ('string', None),
    'h': ('bool', True),
    'i': ('bool', True),
    'ls': ('list[str]', ['a', 'b', 'c']),
    'li': ('list[int]', [42, 43]),
}

other_default_conf = {
    'a': ('int', 3),
}

full_default_conf = default_conf.copy()
full_default_conf['a'] = other_default_conf['a']

parsed_default_conf = {
    key: value
    for key, (type, value)
    in default_conf.items()
}

parsed_conffile = {
    'a': 1,
    'b': 'this is a string',
    'c': True,
    'd': 10,
    'e': None,
    'f': None,
    'g': None,
    'h': False,
    'i': True,
    'ls': ['list', 'of', 'strings'],
    'li': [1, 2, 3, 4],
}


@pytest.fixture
def swh_config(tmp_path):
    # create a temporary folder
    conffile = tmp_path / 'config.ini'
    conf_contents = """[main]
a = 1
b = this is a string
c = true
h = false
ls = list, of, strings
li = 1, 2, 3, 4
"""
    conffile.open('w').write(conf_contents)
    return conffile


@pytest.fixture
def swh_config_unreadable(swh_config):
    # Create an unreadable, proper configuration file
    os.chmod(str(swh_config), 0o000)
    yield swh_config
    # Make the broken perms file readable again to be able to remove them
    os.chmod(str(swh_config), 0o644)


@pytest.fixture
def swh_config_unreadable_dir(swh_config):
    # Create a proper configuration file in an unreadable directory
    perms_broken_dir = swh_config.parent / 'unreadabledir'
    perms_broken_dir.mkdir()
    shutil.move(str(swh_config), str(perms_broken_dir))
    os.chmod(str(perms_broken_dir), 0o000)
    yield perms_broken_dir / swh_config.name
    # Make the broken perms items readable again to be able to remove them
    os.chmod(str(perms_broken_dir), 0o755)


@pytest.fixture
def swh_config_empty(tmp_path):
    # create a temporary folder
    conffile = tmp_path / 'config.ini'
    conffile.touch()
    return conffile


def test_read(swh_config):
    # when
    res = config.read(str(swh_config), default_conf)

    # then
    assert res == parsed_conffile


def test_read_empty_file():
    # when
    res = config.read(None, default_conf)

    # then
    assert res == parsed_default_conf


def test_support_non_existing_conffile(tmp_path):
    # when
    res = config.read(str(tmp_path / 'void.ini'), default_conf)

    # then
    assert res == parsed_default_conf


def test_support_empty_conffile(swh_config_empty):
    # when
    res = config.read(str(swh_config_empty), default_conf)

    # then
    assert res == parsed_default_conf


def test_raise_on_broken_directory_perms(swh_config_unreadable_dir):
    with pytest.raises(PermissionError):
        config.read(str(swh_config_unreadable_dir), default_conf)


def test_raise_on_broken_file_perms(swh_config_unreadable):
    with pytest.raises(PermissionError):
        config.read(str(swh_config_unreadable), default_conf)


def test_merge_default_configs():
    # when
    res = config.merge_default_configs(default_conf, other_default_conf)

    # then
    assert res == full_default_conf


def test_priority_read_nonexist_conf(swh_config):
    noexist = str(swh_config.parent / 'void.ini')
    # when
    res = config.priority_read([noexist, str(swh_config)], default_conf)

    # then
    assert res == parsed_conffile


def test_priority_read_conf_nonexist_empty(swh_config):
    noexist = swh_config.parent / 'void.ini'
    empty = swh_config.parent / 'empty.ini'
    empty.touch()

    # when
    res = config.priority_read([str(p) for p in (
        swh_config, noexist, empty)], default_conf)

    # then
    assert res == parsed_conffile


def test_priority_read_empty_conf_nonexist(swh_config):
    noexist = swh_config.parent / 'void.ini'
    empty = swh_config.parent / 'empty.ini'
    empty.touch()

    # when
    res = config.priority_read([str(p) for p in (
        empty, swh_config, noexist)], default_conf)

    # then
    assert res == parsed_default_conf


def test_swh_config_paths():
    res = config.swh_config_paths('foo/bar.ini')

    assert res == [
        '~/.config/swh/foo/bar.ini',
        '~/.swh/foo/bar.ini',
        '/etc/softwareheritage/foo/bar.ini',
    ]


def test_prepare_folder(tmp_path):
    # given
    conf = {'path1': str(tmp_path / 'path1'),
            'path2': str(tmp_path / 'path2' / 'depth1')}

    # the folders does not exists
    assert not os.path.exists(conf['path1']), "path1 should not exist."
    assert not os.path.exists(conf['path2']), "path2 should not exist."

    # when
    config.prepare_folders(conf, 'path1')

    # path1 exists but not path2
    assert os.path.exists(conf['path1']), "path1 should now exist!"
    assert not os.path.exists(conf['path2']), "path2 should not exist."

    # path1 already exists, skips it but creates path2
    config.prepare_folders(conf, 'path1', 'path2')

    assert os.path.exists(conf['path1']), "path1 should still exist!"
    assert os.path.exists(conf['path2']), "path2 should now exist."


def test_merge_config():
    cfg_a = {
        'a': 42,
        'b': [1, 2, 3],
        'c': None,
        'd': {'gheez': 27},
        'e': {
            'ea': 'Mr. Bungle',
            'eb': None,
            'ec': [11, 12, 13],
            'ed': {'eda': 'Secret Chief 3',
                   'edb': 'Faith No More'},
            'ee': 451,
        },
        'f': 'Janis',
    }
    cfg_b = {
        'a': 43,
        'b': [41, 42, 43],
        'c': 'Tom Waits',
        'd': None,
        'e': {
            'ea': 'Igorrr',
            'ec': [51, 52],
            'ed': {'edb': 'Sleepytime Gorilla Museum',
                   'edc': 'Nils Peter Molvaer'},
        },
        'g': 'Hüsker Dü',
    }

    # merge A, B
    cfg_m = config.merge_configs(cfg_a, cfg_b)
    assert cfg_m == {
        'a': 43,  # b takes precedence
        'b': [41, 42, 43],  # b takes precedence
        'c': 'Tom Waits',  # b takes precedence
        'd': None,  # b['d'] takes precedence (explicit None)
        'e': {
            'ea': 'Igorrr',  # a takes precedence
            'eb': None,  # only in a
            'ec': [51, 52],  # b takes precedence
            'ed': {
                'eda': 'Secret Chief 3',  # only in a
                'edb': 'Sleepytime Gorilla Museum',  # b takes precedence
                'edc': 'Nils Peter Molvaer'},  # only defined in b
            'ee': 451,
        },
        'f': 'Janis',  # only defined in a
        'g': 'Hüsker Dü',  # only defined in b
    }

    # merge B, A
    cfg_m = config.merge_configs(cfg_b, cfg_a)
    assert cfg_m == {
        'a': 42,  # a takes precedence
        'b': [1, 2, 3],  # a takes precedence
        'c': None,  # a takes precedence
        'd': {'gheez': 27},  # a takes precedence
        'e': {
            'ea': 'Mr. Bungle',  # a takes precedence
            'eb': None,  # only defined in a
            'ec': [11, 12, 13],  # a takes precedence
            'ed': {
                'eda': 'Secret Chief 3',  # only in a
                'edb': 'Faith No More',  # a takes precedence
                'edc': 'Nils Peter Molvaer'},  # only in b
            'ee': 451,
        },
        'f': 'Janis',  # only in a
        'g': 'Hüsker Dü',  # only in b
    }


def test_merge_config_type_error():
    for v in (1, 'str', None):
        with pytest.raises(TypeError):
            config.merge_configs(v, {})
        with pytest.raises(TypeError):
            config.merge_configs({}, v)

    for v in (1, 'str'):
        with pytest.raises(TypeError):
            config.merge_configs({'a': v}, {'a': {}})
        with pytest.raises(TypeError):
            config.merge_configs({'a': {}}, {'a': v})



# Copyright (C) 2015-2017  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import collections
import functools
import inspect
import json
import logging
import pickle
import requests
import datetime

from typing import ClassVar, Optional, Type

from deprecated import deprecated
from flask import Flask, Request, Response, request, abort
from .serializers import (decode_response,
                          encode_data_client as encode_data,
                          msgpack_dumps, msgpack_loads, SWHJSONDecoder)

from .negotiation import (Formatter as FormatterBase,
                          Negotiator as NegotiatorBase,
                          negotiate as _negotiate)


logger = logging.getLogger(__name__)


# support for content negotiation

class Negotiator(NegotiatorBase):
    def best_mimetype(self):
        return request.accept_mimetypes.best_match(
            self.accept_mimetypes, 'application/json')

    def _abort(self, status_code, err=None):
        return abort(status_code, err)


def negotiate(formatter_cls, *args, **kwargs):
    return _negotiate(Negotiator, formatter_cls, *args, **kwargs)


class Formatter(FormatterBase):
    def _make_response(self, body, content_type):
        return Response(body, content_type=content_type)


class SWHJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, (datetime.datetime, datetime.date)):
            return obj.isoformat()
        if isinstance(obj, datetime.timedelta):
            return str(obj)
        # Let the base class default method raise the TypeError
        return super().default(obj)


class JSONFormatter(Formatter):
    format = 'json'
    mimetypes = ['application/json']

    def render(self, obj):
        return json.dumps(obj, cls=SWHJSONEncoder)


class MsgpackFormatter(Formatter):
    format = 'msgpack'
    mimetypes = ['application/x-msgpack']

    def render(self, obj):
        return msgpack_dumps(obj)


# base API classes

class RemoteException(Exception):
    pass


def remote_api_endpoint(path):
    def dec(f):
        f._endpoint_path = path
        return f
    return dec


class APIError(Exception):
    """API Error"""
    def __str__(self):
        return ('An unexpected error occurred in the backend: {}'
                .format(self.args))


class MetaRPCClient(type):
    """Metaclass for RPCClient, which adds a method for each endpoint
    of the database it is designed to access.

    See for example :class:`swh.indexer.storage.api.client.RemoteStorage`"""
    def __new__(cls, name, bases, attributes):
        # For each method wrapped with @remote_api_endpoint in an API backend
        # (eg. :class:`swh.indexer.storage.IndexerStorage`), add a new
        # method in RemoteStorage, with the same documentation.
        #
        # Note that, despite the usage of decorator magic (eg. functools.wrap),
        # this never actually calls an IndexerStorage method.
        backend_class = attributes.get('backend_class', None)
        for base in bases:
            if backend_class is not None:
                break
            backend_class = getattr(base, 'backend_class', None)
        if backend_class:
            for (meth_name, meth) in backend_class.__dict__.items():
                if hasattr(meth, '_endpoint_path'):
                    cls.__add_endpoint(meth_name, meth, attributes)
        return super().__new__(cls, name, bases, attributes)

    @staticmethod
    def __add_endpoint(meth_name, meth, attributes):
        wrapped_meth = inspect.unwrap(meth)

        @functools.wraps(meth)  # Copy signature and doc
        def meth_(*args, **kwargs):
            # Match arguments and parameters
            post_data = inspect.getcallargs(
                    wrapped_meth, *args, **kwargs)

            # Remove arguments that should not be passed
            self = post_data.pop('self')
            post_data.pop('cur', None)
            post_data.pop('db', None)

            # Send the request.
            return self.post(meth._endpoint_path, post_data)
        attributes[meth_name] = meth_


class RPCClient(metaclass=MetaRPCClient):
    """Proxy to an internal SWH RPC

    """

    backend_class = None  # type: ClassVar[Optional[type]]
    """For each method of `backend_class` decorated with
    :func:`remote_api_endpoint`, a method with the same prototype and
    docstring will be added to this class. Calls to this new method will
    be translated into HTTP requests to a remote server.

    This backend class will never be instantiated, it only serves as
    a template."""

    api_exception = APIError  # type: ClassVar[Type[Exception]]
    """The exception class to raise in case of communication error with
    the server."""

    def __init__(self, url, api_exception=None,
                 timeout=None, chunk_size=4096, **kwargs):
        if api_exception:
            self.api_exception = api_exception
        base_url = url if url.endswith('/') else url + '/'
        self.url = base_url
        self.session = requests.Session()
        adapter = requests.adapters.HTTPAdapter(
            max_retries=kwargs.get('max_retries', 3),
            pool_connections=kwargs.get('pool_connections', 20),
            pool_maxsize=kwargs.get('pool_maxsize', 100))
        self.session.mount(self.url, adapter)

        self.timeout = timeout
        self.chunk_size = chunk_size

    def _url(self, endpoint):
        return '%s%s' % (self.url, endpoint)

    def raw_verb(self, verb, endpoint, **opts):
        if 'chunk_size' in opts:
            # if the chunk_size argument has been passed, consider the user
            # also wants stream=True, otherwise, what's the point.
            opts['stream'] = True
        if self.timeout and 'timeout' not in opts:
            opts['timeout'] = self.timeout
        try:
            return getattr(self.session, verb)(
                self._url(endpoint),
                **opts
            )
        except requests.exceptions.ConnectionError as e:
            raise self.api_exception(e)

    def post(self, endpoint, data, **opts):
        if isinstance(data, (collections.Iterator, collections.Generator)):
            data = (encode_data(x) for x in data)
        else:
            data = encode_data(data)
        chunk_size = opts.pop('chunk_size', self.chunk_size)
        response = self.raw_verb(
            'post', endpoint, data=data,
            headers={'content-type': 'application/x-msgpack',
                     'accept': 'application/x-msgpack'},
            **opts)
        if opts.get('stream') or \
           response.headers.get('transfer-encoding') == 'chunked':
            return response.iter_content(chunk_size)
        else:
            return self._decode_response(response)

    post_stream = post

    def get(self, endpoint, **opts):
        chunk_size = opts.pop('chunk_size', self.chunk_size)
        response = self.raw_verb(
            'get', endpoint,
            headers={'accept': 'application/x-msgpack'},
            **opts)
        if opts.get('stream') or \
           response.headers.get('transfer-encoding') == 'chunked':
            return response.iter_content(chunk_size)
        else:
            return self._decode_response(response)

    def get_stream(self, endpoint, **opts):
        return self.get(endpoint, stream=True, **opts)

    def _decode_response(self, response):
        if response.status_code == 404:
            return None
        if response.status_code == 500:
            data = decode_response(response)
            if 'exception_pickled' in data:
                raise pickle.loads(data['exception_pickled'])
            else:
                raise RemoteException(data['exception'])

        # XXX: this breaks language-independence and should be
        # replaced by proper unserialization
        if response.status_code == 400:
            raise pickle.loads(decode_response(response))
        elif response.status_code != 200:
            raise RemoteException(
                "Unexpected status code for API request: %s (%s)" % (
                    response.status_code,
                    response.content,
                )
            )
        return decode_response(response)

    def __repr__(self):
        return '<{} url={}>'.format(self.__class__.__name__, self.url)


class BytesRequest(Request):
    """Request with proper escaping of arbitrary byte sequences."""
    encoding = 'utf-8'
    encoding_errors = 'surrogateescape'


ENCODERS = {
    'application/x-msgpack': msgpack_dumps,
    'application/json': json.dumps,
}


def encode_data_server(data, content_type='application/x-msgpack'):
    encoded_data = ENCODERS[content_type](data)
    return Response(
            encoded_data,
            mimetype=content_type,
        )


def decode_request(request):
    content_type = request.mimetype
    data = request.get_data()
    if not data:
        return {}

    if content_type == 'application/x-msgpack':
        r = msgpack_loads(data)
    elif content_type == 'application/json':
        r = json.loads(data, cls=SWHJSONDecoder)
    else:
        raise ValueError('Wrong content type `%s` for API request'
                         % content_type)

    return r


def error_handler(exception, encoder):
    # XXX: this breaks language-independence and should be
    # replaced by proper serialization of errors
    logging.exception(exception)
    response = encoder(pickle.dumps(exception))
    response.status_code = 400
    return response


class RPCServerApp(Flask):
    """For each endpoint of the given `backend_class`, tells app.route to call
    a function that decodes the request and sends it to the backend object
    provided by the factory.

    :param Any backend_class: The class of the backend, which will be
                              analyzed to look for API endpoints.
    :param Callable[[], backend_class] backend_factory: A function with no
                                                        argument that returns
                                                        an instance of
                                                        `backend_class`."""
    request_class = BytesRequest

    def __init__(self, *args, backend_class=None, backend_factory=None,
                 **kwargs):
        super().__init__(*args, **kwargs)

        if backend_class is not None:
            if backend_factory is None:
                raise TypeError('Missing argument backend_factory')
            for (meth_name, meth) in backend_class.__dict__.items():
                if hasattr(meth, '_endpoint_path'):
                    self.__add_endpoint(meth_name, meth, backend_factory)

    def __add_endpoint(self, meth_name, meth, backend_factory):
        from flask import request

        @self.route('/'+meth._endpoint_path, methods=['POST'])
        @functools.wraps(meth)  # Copy signature and doc
        def _f():
            # Call the actual code
            obj_meth = getattr(backend_factory(), meth_name)
            return encode_data_server(obj_meth(**decode_request(request)))


@deprecated(version='0.0.64',
            reason='Use the RPCServerApp instead')
class SWHServerAPIApp(RPCServerApp):
    pass


@deprecated(version='0.0.64',
            reason='Use the MetaRPCClient instead')
class MetaSWHRemoteAPI(MetaRPCClient):
    pass


@deprecated(version='0.0.64',
            reason='Use the RPCClient instead')
class SWHRemoteAPI(RPCClient):
    pass



# Copyright (C) 2015-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import binascii
import datetime
import enum
import json
import logging
import os
import sys
import threading

from contextlib import contextmanager

import psycopg2
import psycopg2.extras


logger = logging.getLogger(__name__)


psycopg2.extras.register_uuid()


def escape(data):
    if data is None:
        return ''
    if isinstance(data, bytes):
        return '\\x%s' % binascii.hexlify(data).decode('ascii')
    elif isinstance(data, str):
        return '"%s"' % data.replace('"', '""')
    elif isinstance(data, datetime.datetime):
        # We escape twice to make sure the string generated by
        # isoformat gets escaped
        return escape(data.isoformat())
    elif isinstance(data, dict):
        return escape(json.dumps(data))
    elif isinstance(data, list):
        return escape("{%s}" % ','.join(escape(d) for d in data))
    elif isinstance(data, psycopg2.extras.Range):
        # We escape twice here too, so that we make sure
        # everything gets passed to copy properly
        return escape(
            '%s%s,%s%s' % (
                '[' if data.lower_inc else '(',
                '-infinity' if data.lower_inf else escape(data.lower),
                'infinity' if data.upper_inf else escape(data.upper),
                ']' if data.upper_inc else ')',
            )
        )
    elif isinstance(data, enum.IntEnum):
        return escape(int(data))
    else:
        # We don't escape here to make sure we pass literals properly
        return str(data)


def typecast_bytea(value, cur):
    if value is not None:
        data = psycopg2.BINARY(value, cur)
        return data.tobytes()


class BaseDb:
    """Base class for swh.*.*Db.

    cf. swh.storage.db.Db, swh.archiver.db.ArchiverDb

    """

    @classmethod
    def adapt_conn(cls, conn):
        """Makes psycopg2 use 'bytes' to decode bytea instead of
        'memoryview', for this connection."""
        cur = conn.cursor()
        cur.execute("SELECT null::bytea, null::bytea[]")
        bytea_oid = cur.description[0][1]
        bytea_array_oid = cur.description[1][1]

        t_bytes = psycopg2.extensions.new_type(
            (bytea_oid,), "bytea", typecast_bytea)
        psycopg2.extensions.register_type(t_bytes, conn)

        t_bytes_array = psycopg2.extensions.new_array_type(
            (bytea_array_oid,), "bytea[]", t_bytes)
        psycopg2.extensions.register_type(t_bytes_array, conn)

    @classmethod
    def connect(cls, *args, **kwargs):
        """factory method to create a DB proxy

        Accepts all arguments of psycopg2.connect; only some specific
        possibilities are reported below.

        Args:
            connstring: libpq2 connection string

        """
        conn = psycopg2.connect(*args, **kwargs)
        return cls(conn)

    @classmethod
    def from_pool(cls, pool):
        conn = pool.getconn()
        return cls(conn, pool=pool)

    def __init__(self, conn, pool=None):
        """create a DB proxy

        Args:
            conn: psycopg2 connection to the SWH DB
            pool: psycopg2 pool of connections

        """
        self.adapt_conn(conn)
        self.conn = conn
        self.pool = pool

    def put_conn(self):
        if self.pool:
            self.pool.putconn(self.conn)

    def cursor(self, cur_arg=None):
        """get a cursor: from cur_arg if given, or a fresh one otherwise

        meant to avoid boilerplate if/then/else in methods that proxy stored
        procedures

        """
        if cur_arg is not None:
            return cur_arg
        else:
            return self.conn.cursor()
    _cursor = cursor  # for bw compat

    @contextmanager
    def transaction(self):
        """context manager to execute within a DB transaction

        Yields:
            a psycopg2 cursor

        """
        with self.conn.cursor() as cur:
            try:
                yield cur
                self.conn.commit()
            except Exception:
                if not self.conn.closed:
                    self.conn.rollback()
                raise

    def copy_to(self, items, tblname, columns,
                cur=None, item_cb=None, default_values={}):
        """Copy items' entries to table tblname with columns information.

        Args:
            items (List[dict]): dictionaries of data to copy over tblname.
            tblname (str): destination table's name.
            columns ([str]): keys to access data in items and also the
              column names in the destination table.
            default_values (dict): dictionary of default values to use when
              inserting entried int the tblname table.
            cur: a db cursor; if not given, a new cursor will be created.
            item_cb (fn): optional function to apply to items's entry.
        """

        read_file, write_file = os.pipe()
        exc_info = None

        def writer():
            nonlocal exc_info
            cursor = self.cursor(cur)
            with open(read_file, 'r') as f:
                try:
                    cursor.copy_expert('COPY %s (%s) FROM STDIN CSV' % (
                        tblname, ', '.join(columns)), f)
                except Exception:
                    # Tell the main thread about the exception
                    exc_info = sys.exc_info()

        write_thread = threading.Thread(target=writer)
        write_thread.start()

        try:
            with open(write_file, 'w') as f:
                for d in items:
                    if item_cb is not None:
                        item_cb(d)
                    line = []
                    for k in columns:
                        try:
                            value = d.get(k, default_values.get(k))
                            line.append(escape(value))
                        except Exception as e:
                            logger.error(
                                'Could not escape value `%r` for column `%s`:'
                                'Received exception: `%s`',
                                value, k, e
                            )
                            raise e from None
                    f.write(','.join(line))
                    f.write('\n')

        finally:
            # No problem bubbling up exceptions, but we still need to make sure
            # we finish copying, even though we're probably going to cancel the
            # transaction.
            write_thread.join()
            if exc_info:
                # postgresql returned an error, let's raise it.
                raise exc_info[1].with_traceback(exc_info[2])

    def mktemp(self, tblname, cur=None):
        self.cursor(cur).execute('SELECT swh_mktemp(%s)', (tblname,))



# This code is a partial and adapted copy of
#   https://github.com/nickstenning/negotiate
#
# Copyright 2012-2013 Nick Stenning
#           2019 The Software Heritage developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

from collections import defaultdict
from decorator import decorator
from inspect import getcallargs

from typing import Any, List, Optional


class FormatterNotFound(Exception):
    pass


class Formatter:
    format = None  # type: Optional[str]
    mimetypes = []  # type: List[Any]

    def __init__(self, request_mimetype=None):
        if request_mimetype is None or request_mimetype not in self.mimetypes:
            try:
                self.response_mimetype = self.mimetypes[0]
            except IndexError:
                raise NotImplementedError(
                    "%s.mimetypes should be a non-empty list" %
                    self.__class__.__name__)
        else:
            self.response_mimetype = request_mimetype

    def configure(self):
        pass

    def render(self, obj):
        raise NotImplementedError(
            "render() should be implemented by Formatter subclasses")

    def __call__(self, obj):
        return self._make_response(
            self.render(obj), content_type=self.response_mimetype)

    def _make_response(self, body, content_type):
        raise NotImplementedError(
            "_make_response() should be implemented by "
            "framework-specific subclasses of Formatter"
        )


class Negotiator:

    def __init__(self, func):
        self.func = func
        self._formatters = []
        self._formatters_by_format = defaultdict(list)
        self._formatters_by_mimetype = defaultdict(list)

    def __call__(self, *args, **kwargs):
        result = self.func(*args, **kwargs)
        format = getcallargs(self.func, *args, **kwargs).get('format')
        mimetype = self.best_mimetype()

        try:
            formatter = self.get_formatter(format, mimetype)
        except FormatterNotFound as e:
            return self._abort(404, str(e))

        return formatter(result)

    def register_formatter(self, formatter, *args, **kwargs):
        self._formatters.append(formatter)
        self._formatters_by_format[formatter.format].append(
            (formatter, args, kwargs))
        for mimetype in formatter.mimetypes:
            self._formatters_by_mimetype[mimetype].append(
                (formatter, args, kwargs))

    def get_formatter(self, format=None, mimetype=None):
        if format is None and mimetype is None:
            raise TypeError(
                "get_formatter expects one of the 'format' or 'mimetype' "
                "kwargs to be set")

        if format is not None:
            try:
                # the first added will be the most specific
                formatter_cls, args, kwargs = (
                    self._formatters_by_format[format][0])
            except IndexError:
                raise FormatterNotFound(
                    "Formatter for format '%s' not found!" % format)
        elif mimetype is not None:
            try:
                # the first added will be the most specific
                formatter_cls, args, kwargs = (
                    self._formatters_by_mimetype[mimetype][0])
            except IndexError:
                raise FormatterNotFound(
                    "Formatter for mimetype '%s' not found!" % mimetype)

        formatter = formatter_cls(request_mimetype=mimetype)
        formatter.configure(*args, **kwargs)
        return formatter

    @property
    def accept_mimetypes(self):
        return [m for f in self._formatters for m in f.mimetypes]

    def best_mimetype(self):
        raise NotImplementedError(
            "best_mimetype() should be implemented in "
            "framework-specific subclasses of Negotiator"
        )

    def _abort(self, status_code, err=None):
        raise NotImplementedError(
            "_abort() should be implemented in framework-specific "
            "subclasses of Negotiator"
        )


def negotiate(negotiator_cls, formatter_cls, *args, **kwargs):
    def _negotiate(f, *args, **kwargs):
        return f.negotiator(*args, **kwargs)

    def decorate(f):
        if not hasattr(f, 'negotiator'):
            f.negotiator = negotiator_cls(f)

        f.negotiator.register_formatter(formatter_cls, *args, **kwargs)
        return decorator(_negotiate, f)

    return decorate



# Copyright (C) 2015-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import unittest

from swh.core import utils


class UtilsLib(unittest.TestCase):

    def test_grouper(self):
        # given
        actual_data = utils.grouper((i for i in range(0, 9)), 2)

        out = []
        for d in actual_data:
            out.append(list(d))  # force generator resolution for checks

        self.assertEqual(out, [[0, 1], [2, 3], [4, 5], [6, 7], [8]])

        # given
        actual_data = utils.grouper((i for i in range(9, 0, -1)), 4)

        out = []
        for d in actual_data:
            out.append(list(d))  # force generator resolution for checks

        self.assertEqual(out, [[9, 8, 7, 6], [5, 4, 3, 2], [1]])

    def test_grouper_with_stop_value(self):
        # given
        actual_data = utils.grouper(((i, i+1) for i in range(0, 9)), 2)

        out = []
        for d in actual_data:
            out.append(list(d))  # force generator resolution for checks

        self.assertEqual(out, [
            [(0, 1), (1, 2)],
            [(2, 3), (3, 4)],
            [(4, 5), (5, 6)],
            [(6, 7), (7, 8)],
            [(8, 9)]])

        # given
        actual_data = utils.grouper((i for i in range(9, 0, -1)), 4)

        out = []
        for d in actual_data:
            out.append(list(d))  # force generator resolution for checks

        self.assertEqual(out, [[9, 8, 7, 6], [5, 4, 3, 2], [1]])

    def test_backslashescape_errors(self):
        raw_data_err = b'abcd\x80'
        with self.assertRaises(UnicodeDecodeError):
            raw_data_err.decode('utf-8', 'strict')

        self.assertEqual(
            raw_data_err.decode('utf-8', 'backslashescape'),
            'abcd\\x80',
        )

        raw_data_ok = b'abcd\xc3\xa9'
        self.assertEqual(
            raw_data_ok.decode('utf-8', 'backslashescape'),
            raw_data_ok.decode('utf-8', 'strict'),
        )

        unicode_data = 'abcdef\u00a3'
        self.assertEqual(
            unicode_data.encode('ascii', 'backslashescape'),
            b'abcdef\\xa3',
        )

    def test_encode_with_unescape(self):
        valid_data = '\\x01020304\\x00'
        valid_data_encoded = b'\x01020304\x00'

        self.assertEqual(
            valid_data_encoded,
            utils.encode_with_unescape(valid_data)
        )

    def test_encode_with_unescape_invalid_escape(self):
        invalid_data = 'test\\abcd'

        with self.assertRaises(ValueError) as exc:
            utils.encode_with_unescape(invalid_data)

        self.assertIn('invalid escape', exc.exception.args[0])
        self.assertIn('position 4', exc.exception.args[0])

    def test_decode_with_escape(self):
        backslashes = b'foo\\bar\\\\baz'
        backslashes_escaped = 'foo\\\\bar\\\\\\\\baz'

        self.assertEqual(
            backslashes_escaped,
            utils.decode_with_escape(backslashes),
        )

        valid_utf8 = b'foo\xc3\xa2'
        valid_utf8_escaped = 'foo\u00e2'

        self.assertEqual(
            valid_utf8_escaped,
            utils.decode_with_escape(valid_utf8),
        )

        invalid_utf8 = b'foo\xa2'
        invalid_utf8_escaped = 'foo\\xa2'

        self.assertEqual(
            invalid_utf8_escaped,
            utils.decode_with_escape(invalid_utf8),
        )

        valid_utf8_nul = b'foo\xc3\xa2\x00'
        valid_utf8_nul_escaped = 'foo\u00e2\\x00'

        self.assertEqual(
            valid_utf8_nul_escaped,
            utils.decode_with_escape(valid_utf8_nul),
        )

    def test_commonname(self):
        # when
        actual_commonname = utils.commonname('/some/where/to/',
                                             '/some/where/to/go/to')
        # then
        self.assertEqual('go/to', actual_commonname)

        # when
        actual_commonname2 = utils.commonname(b'/some/where/to/',
                                              b'/some/where/to/go/to')
        # then
        self.assertEqual(b'go/to', actual_commonname2)



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os.path
import tempfile
import unittest

from hypothesis import strategies, given
import psycopg2
import pytest

from swh.core.db import BaseDb
from .db_testing import (
    SingleDbTestFixture, db_create, db_destroy, db_close,
)


INIT_SQL = '''
create table test_table
(
    i       int,
    txt     text,
    bytes   bytea
);
'''

db_rows = strategies.lists(strategies.tuples(
    strategies.integers(-2147483648, +2147483647),
    strategies.text(
        alphabet=strategies.characters(
            blacklist_categories=['Cs'],  # surrogates
            blacklist_characters=[
                '\x00',  # pgsql does not support the null codepoint
                '\r',  # pgsql normalizes those
            ]
        ),
    ),
    strategies.binary(),
))


@pytest.mark.db
def test_connect():
    db_name = db_create('test-db2', dumps=[])
    try:
        db = BaseDb.connect('dbname=%s' % db_name)
        with db.cursor() as cur:
            cur.execute(INIT_SQL)
            cur.execute("insert into test_table values (1, %s, %s);",
                        ('foo', b'bar'))
            cur.execute("select * from test_table;")
            assert list(cur) == [(1, 'foo', b'bar')]
    finally:
        db_close(db.conn)
        db_destroy(db_name)


@pytest.mark.db
class TestDb(SingleDbTestFixture, unittest.TestCase):
    TEST_DB_NAME = 'test-db'

    @classmethod
    def setUpClass(cls):
        with tempfile.TemporaryDirectory() as td:
            with open(os.path.join(td, 'init.sql'), 'a') as fd:
                fd.write(INIT_SQL)

            cls.TEST_DB_DUMP = os.path.join(td, '*.sql')

            super().setUpClass()

    def setUp(self):
        super().setUp()
        self.db = BaseDb(self.conn)

    def test_initialized(self):
        cur = self.db.cursor()
        cur.execute("insert into test_table values (1, %s, %s);",
                    ('foo', b'bar'))
        cur.execute("select * from test_table;")
        self.assertEqual(list(cur), [(1, 'foo', b'bar')])

    def test_reset_tables(self):
        cur = self.db.cursor()
        cur.execute("insert into test_table values (1, %s, %s);",
                    ('foo', b'bar'))
        self.reset_db_tables('test-db')
        cur.execute("select * from test_table;")
        self.assertEqual(list(cur), [])

    @given(db_rows)
    def test_copy_to(self, data):
        # the table is not reset between runs by hypothesis
        self.reset_db_tables('test-db')

        items = [dict(zip(['i', 'txt', 'bytes'], item)) for item in data]
        self.db.copy_to(items, 'test_table', ['i', 'txt', 'bytes'])

        cur = self.db.cursor()
        cur.execute('select * from test_table;')
        self.assertCountEqual(list(cur), data)

    def test_copy_to_thread_exception(self):
        data = [(2**65, 'foo', b'bar')]

        items = [dict(zip(['i', 'txt', 'bytes'], item)) for item in data]
        with self.assertRaises(psycopg2.errors.NumericValueOutOfRange):
            self.db.copy_to(items, 'test_table', ['i', 'txt', 'bytes'])



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
import msgpack
import json

import pytest

from swh.core.api.asynchronous import RPCServerApp, Response
from swh.core.api.asynchronous import encode_msgpack, decode_request

from swh.core.api.serializers import msgpack_dumps, SWHJSONEncoder


pytest_plugins = ['aiohttp.pytest_plugin', 'pytester']


async def root(request):
    return Response('toor')

STRUCT = {'txt': 'something stupid',
          # 'date': datetime.date(2019, 6, 9),  # not supported
          'datetime': datetime.datetime(2019, 6, 9, 10, 12),
          'timedelta': datetime.timedelta(days=-2, hours=3),
          'int': 42,
          'float': 3.14,
          'subdata': {'int': 42,
                      'datetime': datetime.datetime(2019, 6, 10, 11, 12),
                      },
          'list': [42, datetime.datetime(2019, 9, 10, 11, 12), 'ok'],
          }


async def struct(request):
    return Response(STRUCT)


async def echo(request):
    data = await decode_request(request)
    return Response(data)


async def echo_no_nego(request):
    # let the content negotiation handle the serialization for us...
    data = await decode_request(request)
    ret = encode_msgpack(data)
    return ret


def check_mimetype(src, dst):
    src = src.split(';')[0].strip()
    dst = dst.split(';')[0].strip()
    assert src == dst


@pytest.fixture
def app():
    app = RPCServerApp()
    app.router.add_route('GET', '/', root)
    app.router.add_route('GET', '/struct', struct)
    app.router.add_route('POST', '/echo', echo)
    app.router.add_route('POST', '/echo-no-nego', echo_no_nego)
    return app


async def test_get_simple(app,  aiohttp_client) -> None:
    assert app is not None

    cli = await aiohttp_client(app)
    resp = await cli.get('/')
    assert resp.status == 200
    check_mimetype(resp.headers['Content-Type'], 'application/x-msgpack')
    data = await resp.read()
    value = msgpack.unpackb(data, raw=False)
    assert value == 'toor'


async def test_get_simple_nego(app,  aiohttp_client) -> None:
    cli = await aiohttp_client(app)
    for ctype in ('x-msgpack', 'json'):
        resp = await cli.get('/', headers={'Accept': 'application/%s' % ctype})
        assert resp.status == 200
        check_mimetype(resp.headers['Content-Type'], 'application/%s' % ctype)
        assert (await decode_request(resp)) == 'toor'


async def test_get_struct(app,  aiohttp_client) -> None:
    """Test returned structured from a simple GET data is OK"""
    cli = await aiohttp_client(app)
    resp = await cli.get('/struct')
    assert resp.status == 200
    check_mimetype(resp.headers['Content-Type'], 'application/x-msgpack')
    assert (await decode_request(resp)) == STRUCT


async def test_get_struct_nego(app,  aiohttp_client) -> None:
    """Test returned structured from a simple GET data is OK"""
    cli = await aiohttp_client(app)
    for ctype in ('x-msgpack', 'json'):
        resp = await cli.get('/struct',
                             headers={'Accept': 'application/%s' % ctype})
        assert resp.status == 200
        check_mimetype(resp.headers['Content-Type'], 'application/%s' % ctype)
        assert (await decode_request(resp)) == STRUCT


async def test_post_struct_msgpack(app,  aiohttp_client) -> None:
    """Test that msgpack encoded posted struct data is returned as is"""
    cli = await aiohttp_client(app)
    # simple struct
    resp = await cli.post(
        '/echo',
        headers={'Content-Type': 'application/x-msgpack'},
        data=msgpack_dumps({'toto': 42}))
    assert resp.status == 200
    check_mimetype(resp.headers['Content-Type'], 'application/x-msgpack')
    assert (await decode_request(resp)) == {'toto': 42}
    # complex struct
    resp = await cli.post(
        '/echo',
        headers={'Content-Type': 'application/x-msgpack'},
        data=msgpack_dumps(STRUCT))
    assert resp.status == 200
    check_mimetype(resp.headers['Content-Type'], 'application/x-msgpack')
    assert (await decode_request(resp)) == STRUCT


async def test_post_struct_json(app,  aiohttp_client) -> None:
    """Test that json encoded posted struct data is returned as is"""
    cli = await aiohttp_client(app)

    resp = await cli.post(
        '/echo',
        headers={'Content-Type': 'application/json'},
        data=json.dumps({'toto': 42}, cls=SWHJSONEncoder))
    assert resp.status == 200
    check_mimetype(resp.headers['Content-Type'], 'application/x-msgpack')
    assert (await decode_request(resp)) == {'toto': 42}

    resp = await cli.post(
        '/echo',
        headers={'Content-Type': 'application/json'},
        data=json.dumps(STRUCT, cls=SWHJSONEncoder))
    assert resp.status == 200
    check_mimetype(resp.headers['Content-Type'], 'application/x-msgpack')
    # assert resp.headers['Content-Type'] == 'application/x-msgpack'
    assert (await decode_request(resp)) == STRUCT


async def test_post_struct_nego(app,  aiohttp_client) -> None:
    """Test that json encoded posted struct data is returned as is

    using content negotiation (accept json or msgpack).
    """
    cli = await aiohttp_client(app)

    for ctype in ('x-msgpack', 'json'):
        resp = await cli.post(
            '/echo',
            headers={'Content-Type': 'application/json',
                     'Accept': 'application/%s' % ctype},
            data=json.dumps(STRUCT, cls=SWHJSONEncoder))
        assert resp.status == 200
        check_mimetype(resp.headers['Content-Type'], 'application/%s' % ctype)
        assert (await decode_request(resp)) == STRUCT


async def test_post_struct_no_nego(app,  aiohttp_client) -> None:
    """Test that json encoded posted struct data is returned as msgpack

    when using non-negotiation-compatible handlers.
    """
    cli = await aiohttp_client(app)

    for ctype in ('x-msgpack', 'json'):
        resp = await cli.post(
            '/echo-no-nego',
            headers={'Content-Type': 'application/json',
                     'Accept': 'application/%s' % ctype},
            data=json.dumps(STRUCT, cls=SWHJSONEncoder))
        assert resp.status == 200
        check_mimetype(resp.headers['Content-Type'], 'application/x-msgpack')
        assert (await decode_request(resp)) == STRUCT



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import unittest

import requests_mock
from werkzeug.wrappers import BaseResponse
from werkzeug.test import Client as WerkzeugTestClient

from swh.core.api import (
        error_handler, encode_data_server,
        remote_api_endpoint, RPCClient, RPCServerApp)


class ApiTest(unittest.TestCase):
    def test_server(self):
        testcase = self
        nb_endpoint_calls = 0

        class TestStorage:
            @remote_api_endpoint('test_endpoint_url')
            def test_endpoint(self, test_data, db=None, cur=None):
                nonlocal nb_endpoint_calls
                nb_endpoint_calls += 1

                testcase.assertEqual(test_data, 'spam')
                return 'egg'

        app = RPCServerApp('testapp',
                           backend_class=TestStorage,
                           backend_factory=lambda: TestStorage())

        @app.errorhandler(Exception)
        def my_error_handler(exception):
            return error_handler(exception, encode_data_server)

        client = WerkzeugTestClient(app, BaseResponse)
        res = client.post('/test_endpoint_url',
                          headers={'Content-Type': 'application/x-msgpack'},
                          data=b'\x81\xa9test_data\xa4spam')

        self.assertEqual(nb_endpoint_calls, 1)
        self.assertEqual(b''.join(res.response), b'\xa3egg')

    def test_client(self):
        class TestStorage:
            @remote_api_endpoint('test_endpoint_url')
            def test_endpoint(self, test_data, db=None, cur=None):
                pass

        nb_http_calls = 0

        def callback(request, context):
            nonlocal nb_http_calls
            nb_http_calls += 1
            self.assertEqual(request.headers['Content-Type'],
                             'application/x-msgpack')
            self.assertEqual(request.body, b'\x81\xa9test_data\xa4spam')
            context.headers['Content-Type'] = 'application/x-msgpack'
            context.content = b'\xa3egg'
            return b'\xa3egg'

        adapter = requests_mock.Adapter()
        adapter.register_uri('POST',
                             'mock://example.com/test_endpoint_url',
                             content=callback)

        class Testclient(RPCClient):
            backend_class = TestStorage

            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                # we need to mount the mock adapter on the base url to override
                # RPCClient's mechanism that also mounts an HTTPAdapter
                # (for configuration purpose)
                self.session.mount('mock://example.com/', adapter)

        c = Testclient(url='mock://example.com/')
        res = c.test_endpoint('spam')

        self.assertEqual(nb_http_calls, 1)
        self.assertEqual(res, 'egg')



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import requests

from os import path

from swh.core.pytest_plugin import requests_mock_datadir_factory


def test_get_response_cb_with_visits_nominal(requests_mock_datadir_visits):
    response = requests.get('https://example.com/file.json')
    assert response.ok
    assert response.json() == {'hello': 'you'}

    response = requests.get('https://example.com/file.json')
    assert response.ok
    assert response.json() == {'hello': 'world'}

    response = requests.get('https://example.com/file.json')
    assert not response.ok
    assert response.status_code == 404


def test_get_response_cb_with_visits(requests_mock_datadir_visits):
    response = requests.get('https://example.com/file.json')
    assert response.ok
    assert response.json() == {'hello': 'you'}

    response = requests.get('https://example.com/other.json')
    assert response.ok
    assert response.json() == "foobar"

    response = requests.get('https://example.com/file.json')
    assert response.ok
    assert response.json() == {'hello': 'world'}

    response = requests.get('https://example.com/other.json')
    assert not response.ok
    assert response.status_code == 404

    response = requests.get('https://example.com/file.json')
    assert not response.ok
    assert response.status_code == 404


def test_get_response_cb_no_visit(requests_mock_datadir):
    response = requests.get('https://example.com/file.json')
    assert response.ok
    assert response.json() == {'hello': 'you'}

    response = requests.get('https://example.com/file.json')
    assert response.ok
    assert response.json() == {'hello': 'you'}


def test_get_response_cb_query_params(requests_mock_datadir):
    response = requests.get('https://example.com/file.json?toto=42')
    assert not response.ok
    assert response.status_code == 404

    response = requests.get(
        'https://example.com/file.json?name=doe&firstname=jane')
    assert response.ok
    assert response.json() == {'hello': 'jane doe'}


requests_mock_datadir_ignore = requests_mock_datadir_factory(
    ignore_urls=['https://example.com/file.json'],
    has_multi_visit=False,
)


def test_get_response_cb_ignore_url(requests_mock_datadir_ignore):
    response = requests.get('https://example.com/file.json')
    assert not response.ok
    assert response.status_code == 404


requests_mock_datadir_ignore_and_visit = requests_mock_datadir_factory(
    ignore_urls=['https://example.com/file.json'],
    has_multi_visit=True,
)


def test_get_response_cb_ignore_url_with_visit(
        requests_mock_datadir_ignore_and_visit):
    response = requests.get('https://example.com/file.json')
    assert not response.ok
    assert response.status_code == 404

    response = requests.get('https://example.com/file.json')
    assert not response.ok
    assert response.status_code == 404


def test_data_dir(datadir):
    expected_datadir = path.join(path.abspath(path.dirname(__file__)), 'data')
    assert datadir == expected_datadir



# Copyright (C) 2015  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import configparser
import logging
import os
import yaml
from itertools import chain
from copy import deepcopy

from typing import Any, Dict, Optional, Tuple


logger = logging.getLogger(__name__)


SWH_CONFIG_DIRECTORIES = [
    '~/.config/swh',
    '~/.swh',
    '/etc/softwareheritage',
]

SWH_GLOBAL_CONFIG = 'global.ini'

SWH_DEFAULT_GLOBAL_CONFIG = {
    'content_size_limit': ('int', 100 * 1024 * 1024),
    'log_db': ('str', 'dbname=softwareheritage-log'),
}

SWH_CONFIG_EXTENSIONS = [
    '.yml',
    '.ini',
]

# conversion per type
_map_convert_fn = {
    'int': int,
    'bool': lambda x: x.lower() == 'true',
    'list[str]': lambda x: [value.strip() for value in x.split(',')],
    'list[int]': lambda x: [int(value.strip()) for value in x.split(',')],
}

_map_check_fn = {
    'int': lambda x: isinstance(x, int),
    'bool': lambda x: isinstance(x, bool),
    'list[str]': lambda x: (isinstance(x, list) and
                            all(isinstance(y, str) for y in x)),
    'list[int]': lambda x: (isinstance(x, list) and
                            all(isinstance(y, int) for y in x)),
}


def exists_accessible(file):
    """Check whether a file exists, and is accessible.

    Returns:
        True if the file exists and is accessible
        False if the file does not exist

    Raises:
        PermissionError if the file cannot be read.
    """

    try:
        os.stat(file)
    except PermissionError:
        raise
    except FileNotFoundError:
        return False
    else:
        if os.access(file, os.R_OK):
            return True
        else:
            raise PermissionError("Permission denied: %r" % file)


def config_basepath(config_path):
    """Return the base path of a configuration file"""
    if config_path.endswith(('.ini', '.yml')):
        return config_path[:-4]

    return config_path


def read_raw_config(base_config_path):
    """Read the raw config corresponding to base_config_path.

    Can read yml or ini files.
    """
    yml_file = base_config_path + '.yml'
    if exists_accessible(yml_file):
        logger.info('Loading config file %s', yml_file)
        with open(yml_file) as f:
            return yaml.safe_load(f)

    ini_file = base_config_path + '.ini'
    if exists_accessible(ini_file):
        config = configparser.ConfigParser()
        config.read(ini_file)
        if 'main' in config._sections:
            logger.info('Loading config file %s', ini_file)
            return config._sections['main']
        else:
            logger.warning('Ignoring config file %s (no [main] section)',
                           ini_file)

    return {}


def config_exists(config_path):
    """Check whether the given config exists"""
    basepath = config_basepath(config_path)
    return any(exists_accessible(basepath + extension)
               for extension in SWH_CONFIG_EXTENSIONS)


def read(conf_file=None, default_conf=None):
    """Read the user's configuration file.

    Fill in the gap using `default_conf`.  `default_conf` is similar to this::

        DEFAULT_CONF = {
            'a': ('str', '/tmp/swh-loader-git/log'),
            'b': ('str', 'dbname=swhloadergit')
            'c': ('bool', true)
            'e': ('bool', None)
            'd': ('int', 10)
        }

    If conf_file is None, return the default config.

    """
    conf = {}

    if conf_file:
        base_config_path = config_basepath(os.path.expanduser(conf_file))
        conf = read_raw_config(base_config_path)

    if not default_conf:
        default_conf = {}

    # remaining missing default configuration key are set
    # also type conversion is enforced for underneath layer
    for key in default_conf:
        nature_type, default_value = default_conf[key]
        val = conf.get(key, None)
        if val is None:  # fallback to default value
            conf[key] = default_value
        elif not _map_check_fn.get(nature_type, lambda x: True)(val):
            # value present but not in the proper format, force type conversion
            conf[key] = _map_convert_fn.get(nature_type, lambda x: x)(val)

    return conf


def priority_read(conf_filenames, default_conf=None):
    """Try reading the configuration files from conf_filenames, in order,
       and return the configuration from the first one that exists.

       default_conf has the same specification as it does in read.
    """

    # Try all the files in order
    for filename in conf_filenames:
        full_filename = os.path.expanduser(filename)
        if config_exists(full_filename):
            return read(full_filename, default_conf)

    # Else, return the default configuration
    return read(None, default_conf)


def merge_default_configs(base_config, *other_configs):
    """Merge several default config dictionaries, from left to right"""
    full_config = base_config.copy()

    for config in other_configs:
        full_config.update(config)

    return full_config


def merge_configs(base, other):
    """Merge two config dictionaries

    This does merge config dicts recursively, with the rules, for every value
    of the dicts (with 'val' not being a dict):

    - None + type -> type
    - type + None -> None
    - dict + dict -> dict (merged)
    - val + dict -> TypeError
    - dict + val -> TypeError
    - val + val -> val (other)

    for instance:

    >>> d1 = {
    ...   'key1': {
    ...     'skey1': 'value1',
    ...     'skey2': {'sskey1': 'value2'},
    ...   },
    ...   'key2': 'value3',
    ... }

    with

    >>> d2 = {
    ...   'key1': {
    ...     'skey1': 'value4',
    ...     'skey2': {'sskey2': 'value5'},
    ...   },
    ...   'key3': 'value6',
    ... }

    will give:

    >>> d3 = {
    ...   'key1': {
    ...     'skey1': 'value4',  # <-- note this
    ...     'skey2': {
    ...       'sskey1': 'value2',
    ...       'sskey2': 'value5',
    ...     },
    ...   },
    ...   'key2': 'value3',
    ...   'key3': 'value6',
    ... }
    >>> assert merge_configs(d1, d2) == d3

    Note that no type checking is done for anything but dicts.
    """
    if not isinstance(base, dict) or not isinstance(other, dict):
        raise TypeError(
            'Cannot merge a %s with a %s' % (type(base), type(other)))

    output = {}
    allkeys = set(chain(base.keys(), other.keys()))
    for k in allkeys:
        vb = base.get(k)
        vo = other.get(k)

        if isinstance(vo, dict):
            output[k] = merge_configs(vb is not None and vb or {}, vo)
        elif isinstance(vb, dict) and k in other and other[k] is not None:
            output[k] = merge_configs(vb, vo is not None and vo or {})
        elif k in other:
            output[k] = deepcopy(vo)
        else:
            output[k] = deepcopy(vb)

    return output


def swh_config_paths(base_filename):
    """Return the Software Heritage specific configuration paths for the given
       filename."""

    return [os.path.join(dirname, base_filename)
            for dirname in SWH_CONFIG_DIRECTORIES]


def prepare_folders(conf, *keys):
    """Prepare the folder mentioned in config under keys.
    """
    def makedir(folder):
        if not os.path.exists(folder):
            os.makedirs(folder)

    for key in keys:
        makedir(conf[key])


def load_global_config():
    """Load the global Software Heritage config"""

    return priority_read(
        swh_config_paths(SWH_GLOBAL_CONFIG),
        SWH_DEFAULT_GLOBAL_CONFIG,
    )


def load_named_config(name, default_conf=None, global_conf=True):
    """Load the config named `name` from the Software Heritage
       configuration paths.

       If global_conf is True (default), read the global configuration
       too.
    """

    conf = {}

    if global_conf:
        conf.update(load_global_config())

    conf.update(priority_read(swh_config_paths(name), default_conf))

    return conf


class SWHConfig:
    """Mixin to add configuration parsing abilities to classes

    The class should override the class attributes:
        - DEFAULT_CONFIG (default configuration to be parsed)
        - CONFIG_BASE_FILENAME (the filename of the configuration to be used)

    This class defines one classmethod, parse_config_file, which
    parses a configuration file using the default config as set in the
    class attribute.

    """

    DEFAULT_CONFIG = {}  # type: Dict[str, Tuple[str, Any]]
    CONFIG_BASE_FILENAME = ''  # type: Optional[str]

    @classmethod
    def parse_config_file(cls, base_filename=None, config_filename=None,
                          additional_configs=None, global_config=True):
        """Parse the configuration file associated to the current class.

        By default, parse_config_file will load the configuration
        cls.CONFIG_BASE_FILENAME from one of the Software Heritage
        configuration directories, in order, unless it is overridden
        by base_filename or config_filename (which shortcuts the file
        lookup completely).

        Args:
            - base_filename (str): overrides the default
              cls.CONFIG_BASE_FILENAME
            - config_filename (str): sets the file to parse instead of
              the defaults set from cls.CONFIG_BASE_FILENAME
            - additional_configs: (list of default configuration dicts)
              allows to override or extend the configuration set in
              cls.DEFAULT_CONFIG.
            - global_config (bool): Load the global configuration (default:
              True)
        """

        if config_filename:
            config_filenames = [config_filename]
        elif 'SWH_CONFIG_FILENAME' in os.environ:
            config_filenames = [os.environ['SWH_CONFIG_FILENAME']]
        else:
            if not base_filename:
                base_filename = cls.CONFIG_BASE_FILENAME
            config_filenames = swh_config_paths(base_filename)
        if not additional_configs:
            additional_configs = []

        full_default_config = merge_default_configs(cls.DEFAULT_CONFIG,
                                                    *additional_configs)

        config = {}
        if global_config:
            config = load_global_config()

        config.update(priority_read(config_filenames, full_default_config))

        return config



# Copyright (C) 2015  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
import logging

from systemd.journal import JournalHandler as _JournalHandler, send

try:
    from celery import current_task
except ImportError:
    current_task = None


EXTRA_LOGDATA_PREFIX = 'swh_'


def db_level_of_py_level(lvl):
    """convert a log level of the logging module to a log level suitable for the
    logging Postgres DB

    """
    return logging.getLevelName(lvl).lower()


def get_extra_data(record, task_args=True):
    """Get the extra data to insert to the database from the logging record"""
    log_data = record.__dict__

    extra_data = {k[len(EXTRA_LOGDATA_PREFIX):]: v
                  for k, v in log_data.items()
                  if k.startswith(EXTRA_LOGDATA_PREFIX)}

    args = log_data.get('args')
    if args:
        extra_data['logging_args'] = args

    # Retrieve Celery task info
    if current_task and current_task.request:
        extra_data['task'] = {
            'id': current_task.request.id,
            'name': current_task.name,
        }
        if task_args:
            extra_data['task'].update({
                'kwargs': current_task.request.kwargs,
                'args': current_task.request.args,
            })

    return extra_data


def flatten(data, separator='_'):
    """Flatten the data dictionary into a flat structure"""
    def inner_flatten(data, prefix):
        if isinstance(data, dict):
            for key, value in data.items():
                yield from inner_flatten(value, prefix + [key])
        elif isinstance(data, (list, tuple)):
            for key, value in enumerate(data):
                yield from inner_flatten(value, prefix + [str(key)])
        else:
            yield prefix, data

    for path, value in inner_flatten(data, []):
        yield separator.join(path), value


def stringify(value):
    """Convert value to string"""
    if isinstance(value, datetime.datetime):
        return value.isoformat()

    return str(value)


class JournalHandler(_JournalHandler):
    def emit(self, record):
        """Write `record` as a journal event.

        MESSAGE is taken from the message provided by the user, and PRIORITY,
        LOGGER, THREAD_NAME, CODE_{FILE,LINE,FUNC} fields are appended
        automatically. In addition, record.MESSAGE_ID will be used if present.
        """
        try:
            extra_data = flatten(get_extra_data(record, task_args=False))
            extra_data = {
                (EXTRA_LOGDATA_PREFIX + key).upper(): stringify(value)
                for key, value in extra_data
            }
            msg = self.format(record)
            pri = self.mapPriority(record.levelno)
            send(msg,
                 PRIORITY=format(pri),
                 LOGGER=record.name,
                 THREAD_NAME=record.threadName,
                 CODE_FILE=record.pathname,
                 CODE_LINE=record.lineno,
                 CODE_FUNC=record.funcName,
                 **extra_data)
        except Exception:
            self.handleError(record)



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from datetime import datetime
import logging
import pytz
import inspect

from unittest.mock import patch

from swh.core import logger


def lineno():
    """Returns the current line number in our program."""
    return inspect.currentframe().f_back.f_lineno


def test_db_level():
    assert logger.db_level_of_py_level(10) == 'debug'
    assert logger.db_level_of_py_level(20) == 'info'
    assert logger.db_level_of_py_level(30) == 'warning'
    assert logger.db_level_of_py_level(40) == 'error'
    assert logger.db_level_of_py_level(50) == 'critical'


def test_flatten_scalar():
    assert list(logger.flatten('')) == [('', '')]
    assert list(logger.flatten('toto')) == [('', 'toto')]

    assert list(logger.flatten(10)) == [('', 10)]
    assert list(logger.flatten(10.5)) == [('', 10.5)]


def test_flatten_list():
    assert list(logger.flatten([])) == []
    assert list(logger.flatten([1])) == [('0', 1)]

    assert list(logger.flatten([1, 2, ['a', 'b']])) == [
        ('0', 1),
        ('1', 2),
        ('2_0', 'a'),
        ('2_1', 'b'),
    ]

    assert list(logger.flatten([1, 2, ['a', ('x', 1)]])) == [
        ('0', 1),
        ('1', 2),
        ('2_0', 'a'),
        ('2_1_0', 'x'),
        ('2_1_1', 1),
    ]


def test_flatten_dict():
    assert list(logger.flatten({})) == []
    assert list(logger.flatten({'a': 1})) == [('a', 1)]

    assert sorted(logger.flatten({'a': 1,
                                  'b': (2, 3,),
                                  'c': {'d': 4, 'e': 'f'}})) == [
        ('a', 1),
        ('b_0', 2),
        ('b_1', 3),
        ('c_d', 4),
        ('c_e', 'f'),
    ]


def test_stringify():
    assert logger.stringify(None) == 'None'
    assert logger.stringify(123) == '123'
    assert logger.stringify('abc') == 'abc'

    date = datetime(2019, 9, 1, 16, 32)
    assert logger.stringify(date) == '2019-09-01T16:32:00'

    tzdate = datetime(2019, 9, 1, 16, 32, tzinfo=pytz.utc)
    assert logger.stringify(tzdate) == '2019-09-01T16:32:00+00:00'


@patch('swh.core.logger.send')
def test_journal_handler(send):
    log = logging.getLogger('test_logger')
    log.addHandler(logger.JournalHandler())
    log.setLevel(logging.DEBUG)

    _, ln = log.info('hello world'), lineno()

    send.assert_called_with(
        'hello world',
        CODE_FILE=__file__,
        CODE_FUNC='test_journal_handler',
        CODE_LINE=ln,
        LOGGER='test_logger',
        PRIORITY='6',
        THREAD_NAME='MainThread')


@patch('swh.core.logger.send')
def test_journal_handler_w_data(send):
    log = logging.getLogger('test_logger')
    log.addHandler(logger.JournalHandler())
    log.setLevel(logging.DEBUG)

    _, ln = log.debug('something cool %s', ['with', {'extra': 'data'}]), lineno()  # noqa

    send.assert_called_with(
        "something cool ['with', {'extra': 'data'}]",
        CODE_FILE=__file__,
        CODE_FUNC='test_journal_handler_w_data',
        CODE_LINE=ln,
        LOGGER='test_logger',
        PRIORITY='7',
        THREAD_NAME='MainThread',
        SWH_LOGGING_ARGS_0_0='with',
        SWH_LOGGING_ARGS_0_1_EXTRA='data'
    )



# Copyright (C) 2015-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
import json
import unittest
from uuid import UUID

import arrow

from swh.core.api.serializers import (
    SWHJSONDecoder,
    SWHJSONEncoder,
    msgpack_dumps,
    msgpack_loads
)


class Serializers(unittest.TestCase):
    def setUp(self):
        self.tz = datetime.timezone(datetime.timedelta(minutes=118))

        self.data = {
            'bytes': b'123456789\x99\xaf\xff\x00\x12',
            'datetime_naive': datetime.datetime(2015, 1, 1, 12, 4, 42, 231455),
            'datetime_tz': datetime.datetime(2015, 3, 4, 18, 25, 13, 1234,
                                             tzinfo=self.tz),
            'datetime_utc': datetime.datetime(2015, 3, 4, 18, 25, 13, 1234,
                                              tzinfo=datetime.timezone.utc),
            'datetime_delta': datetime.timedelta(64),
            'arrow_date': arrow.get('2018-04-25T16:17:53.533672+00:00'),
            'swhtype': 'fake',
            'swh_dict': {'swhtype': 42, 'd': 'test'},
            'random_dict': {'swhtype': 43},
            'uuid': UUID('cdd8f804-9db6-40c3-93ab-5955d3836234'),
        }

        self.encoded_data = {
            'bytes': {'swhtype': 'bytes', 'd': 'F)}kWH8wXmIhn8j01^'},
            'datetime_naive': {'swhtype': 'datetime',
                               'd': '2015-01-01T12:04:42.231455'},
            'datetime_tz': {'swhtype': 'datetime',
                            'd': '2015-03-04T18:25:13.001234+01:58'},
            'datetime_utc': {'swhtype': 'datetime',
                             'd': '2015-03-04T18:25:13.001234+00:00'},
            'datetime_delta': {'swhtype': 'timedelta',
                               'd': {'days': 64, 'seconds': 0,
                                     'microseconds': 0}},
            'arrow_date': {'swhtype': 'arrow',
                           'd': '2018-04-25T16:17:53.533672+00:00'},
            'swhtype': 'fake',
            'swh_dict': {'swhtype': 42, 'd': 'test'},
            'random_dict': {'swhtype': 43},
            'uuid': {'swhtype': 'uuid',
                     'd': 'cdd8f804-9db6-40c3-93ab-5955d3836234'},
        }

        self.generator = (i for i in range(5))
        self.gen_lst = list(range(5))

    def test_round_trip_json(self):
        data = json.dumps(self.data, cls=SWHJSONEncoder)
        self.assertEqual(self.data, json.loads(data, cls=SWHJSONDecoder))

    def test_encode_swh_json(self):
        data = json.dumps(self.data, cls=SWHJSONEncoder)
        self.assertEqual(self.encoded_data, json.loads(data))

    def test_round_trip_msgpack(self):
        data = msgpack_dumps(self.data)
        self.assertEqual(self.data, msgpack_loads(data))

    def test_generator_json(self):
        data = json.dumps(self.generator, cls=SWHJSONEncoder)
        self.assertEqual(self.gen_lst, json.loads(data, cls=SWHJSONDecoder))

    def test_generator_msgpack(self):
        data = msgpack_dumps(self.generator)
        self.assertEqual(self.gen_lst, msgpack_loads(data))



# Copyright (C) 2015-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import base64
import datetime
from json import JSONDecoder, JSONEncoder
import types
from uuid import UUID

import arrow
import dateutil.parser
import msgpack


def encode_data_client(data):
    try:
        return msgpack_dumps(data)
    except OverflowError as e:
        raise ValueError('Limits were reached. Please, check your input.\n' +
                         str(e))


def decode_response(response):
    content_type = response.headers['content-type']

    if content_type.startswith('application/x-msgpack'):
        r = msgpack_loads(response.content)
    elif content_type.startswith('application/json'):
        r = response.json(cls=SWHJSONDecoder)
    else:
        raise ValueError('Wrong content type `%s` for API response'
                         % content_type)

    return r


class SWHJSONEncoder(JSONEncoder):
    """JSON encoder for data structures generated by Software Heritage.

    This JSON encoder extends the default Python JSON encoder and adds
    awareness for the following specific types:

    - bytes (get encoded as a Base85 string);
    - datetime.datetime (get encoded as an ISO8601 string).

    Non-standard types get encoded as a a dictionary with two keys:

    - swhtype with value 'bytes' or 'datetime';
    - d containing the encoded value.

    SWHJSONEncoder also encodes arbitrary iterables as a list
    (allowing serialization of generators).

    Caveats: Limitations in the JSONEncoder extension mechanism
    prevent us from "escaping" dictionaries that only contain the
    swhtype and d keys, and therefore arbitrary data structures can't
    be round-tripped through SWHJSONEncoder and SWHJSONDecoder.

    """

    def default(self, o):
        if isinstance(o, bytes):
            return {
                'swhtype': 'bytes',
                'd': base64.b85encode(o).decode('ascii'),
            }
        elif isinstance(o, datetime.datetime):
            return {
                'swhtype': 'datetime',
                'd': o.isoformat(),
            }
        elif isinstance(o, UUID):
            return {
                'swhtype': 'uuid',
                'd': str(o),
            }
        elif isinstance(o, datetime.timedelta):
            return {
                'swhtype': 'timedelta',
                'd': {
                    'days': o.days,
                    'seconds': o.seconds,
                    'microseconds': o.microseconds,
                },
            }
        elif isinstance(o, arrow.Arrow):
            return {
                'swhtype': 'arrow',
                'd': o.isoformat(),
            }
        try:
            return super().default(o)
        except TypeError as e:
            try:
                iterable = iter(o)
            except TypeError:
                raise e from None
            else:
                return list(iterable)


class SWHJSONDecoder(JSONDecoder):
    """JSON decoder for data structures encoded with SWHJSONEncoder.

    This JSON decoder extends the default Python JSON decoder,
    allowing the decoding of:

    - bytes (encoded as a Base85 string);
    - datetime.datetime (encoded as an ISO8601 string).

    Non-standard types must be encoded as a a dictionary with exactly
    two keys:

    - swhtype with value 'bytes' or 'datetime';
    - d containing the encoded value.

    To limit the impact our encoding, if the swhtype key doesn't
    contain a known value, the dictionary is decoded as-is.

    """
    def decode_data(self, o):
        if isinstance(o, dict):
            if set(o.keys()) == {'d', 'swhtype'}:
                datatype = o['swhtype']
                if datatype == 'bytes':
                    return base64.b85decode(o['d'])
                elif datatype == 'datetime':
                    return dateutil.parser.parse(o['d'])
                elif datatype == 'uuid':
                    return UUID(o['d'])
                elif datatype == 'timedelta':
                    return datetime.timedelta(**o['d'])
                elif datatype == 'arrow':
                    return arrow.get(o['d'])
            return {key: self.decode_data(value) for key, value in o.items()}
        if isinstance(o, list):
            return [self.decode_data(value) for value in o]
        else:
            return o

    def raw_decode(self, s, idx=0):
        data, index = super().raw_decode(s, idx)
        return self.decode_data(data), index


def msgpack_dumps(data):
    """Write data as a msgpack stream"""
    def encode_types(obj):
        if isinstance(obj, datetime.datetime):
            return {b'__datetime__': True, b's': obj.isoformat()}
        if isinstance(obj, types.GeneratorType):
            return list(obj)
        if isinstance(obj, UUID):
            return {b'__uuid__': True, b's': str(obj)}
        if isinstance(obj, datetime.timedelta):
            return {
                b'__timedelta__': True,
                b's': {
                    'days': obj.days,
                    'seconds': obj.seconds,
                    'microseconds': obj.microseconds,
                },
            }
        if isinstance(obj, arrow.Arrow):
            return {b'__arrow__': True, b's': obj.isoformat()}
        return obj

    return msgpack.packb(data, use_bin_type=True, default=encode_types)


def msgpack_loads(data):
    """Read data as a msgpack stream"""
    def decode_types(obj):
        if b'__datetime__' in obj and obj[b'__datetime__']:
            return dateutil.parser.parse(obj[b's'])
        if b'__uuid__' in obj and obj[b'__uuid__']:
            return UUID(obj[b's'])
        if b'__timedelta__' in obj and obj[b'__timedelta__']:
            return datetime.timedelta(**obj[b's'])
        if b'__arrow__' in obj and obj[b'__arrow__']:
            return arrow.get(obj[b's'])
        return obj

    try:
        return msgpack.unpackb(data, raw=False,
                               object_hook=decode_types)
    except TypeError:  # msgpack < 0.5.2
        return msgpack.unpackb(data, encoding='utf-8',
                               object_hook=decode_types)



#

from click.testing import CliRunner

from swh.core.cli import swh as swhmain
from swh.core.cli.db import db as swhdb


help_msg = '''Usage: swh [OPTIONS] COMMAND [ARGS]...

  Command line interface for Software Heritage.

Options:
  -l, --log-level [NOTSET|DEBUG|INFO|WARNING|ERROR|CRITICAL]
                                  Log level (default to INFO)
  -h, --help                      Show this message and exit.

Commands:
  db  Software Heritage database generic tools.
'''


def test_swh_help():
    swhmain.add_command(swhdb)
    runner = CliRunner()
    result = runner.invoke(swhmain, ['-h'])
    assert result.exit_code == 0
    assert result.output == help_msg


help_db_msg = '''Usage: swh db [OPTIONS] COMMAND [ARGS]...

  Software Heritage database generic tools.

Options:
  -C, --config-file FILE  Configuration file.
  -h, --help              Show this message and exit.

Commands:
  init  Initialize the database for every Software Heritage module found in...
'''


def test_swh_db_help():
    swhmain.add_command(swhdb)
    runner = CliRunner()
    result = runner.invoke(swhmain, ['db', '-h'])
    assert result.exit_code == 0
    assert result.output == help_db_msg



#!/usr/bin/env python3
# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import glob
import logging
from os import path, environ
import subprocess
import warnings

warnings.filterwarnings("ignore")  # noqa prevent psycopg from telling us sh*t

import click

from swh.core.cli import CONTEXT_SETTINGS
from swh.core.config import read as config_read

logger = logging.getLogger(__name__)


@click.group(name="db", context_settings=CONTEXT_SETTINGS)
@click.option("--config-file", "-C", default=None,
              type=click.Path(exists=True, dir_okay=False),
              help="Configuration file.")
@click.pass_context
def db(ctx, config_file):
    """Software Heritage database generic tools.
    """
    ctx.ensure_object(dict)
    if config_file is None:
        config_file = environ.get('SWH_CONFIG_FILENAME')
    cfg = config_read(config_file)
    ctx.obj["config"] = cfg


@db.command(name="init", context_settings=CONTEXT_SETTINGS)
@click.pass_context
def init(ctx):
    """Initialize the database for every Software Heritage module found in the
    configuration file. For every configuration section in the config file
    that:

    1. has the name of an existing swh package,
    2. has credentials for a local db access,

    it will run the initialization scripts from the swh package against the
    given database.

    Example for the config file::

    \b
      storage:
        cls: local
        args:
          db: postgresql:///?service=swh-storage
      objstorage:
        cls: remote
        args:
          url: http://swh-objstorage:5003/

    the command:

      swh db -C /path/to/config.yml init

    will initialize the database for the `storage` section using initialization
    scripts from the `swh.storage` package.
    """

    for modname, cfg in ctx.obj["config"].items():
        if cfg.get("cls") == "local" and cfg.get("args"):
            try:
                sqlfiles = get_sql_for_package(modname)
            except click.BadParameter:
                logger.info(
                    "Failed to load/find sql initialization files for %s",
                    modname)

            if sqlfiles:
                conninfo = cfg["args"]["db"]
                for sqlfile in sqlfiles:
                    subprocess.check_call(
                        [
                            "psql",
                            "--quiet",
                            "--no-psqlrc",
                            "-v",
                            "ON_ERROR_STOP=1",
                            "-d",
                            conninfo,
                            "-f",
                            sqlfile,
                        ]
                    )


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument('module', nargs=-1, required=True)
@click.option('--db-name', '-d', help='Database name.',
              default='softwareheritage-dev', show_default=True)
@click.option('--create-db/--no-create-db', '-C',
              help='Attempt to create the database.',
              default=False)
def db_init(module, db_name, create_db):
    """Initialise a database for the Software Heritage <module>.  By
    default, does not attempt to create the database.

    Example:

      swh db-init -d swh-test storage

    If you want to specify non-default postgresql connection parameters,
    please provide them using standard environment variables.
    See psql(1) man page (section ENVIRONMENTS) for details.

    Example:

      PGPORT=5434 swh db-init indexer

    """
    # put import statements here so we can keep startup time of the main swh
    # command as short as possible
    from swh.core.db.tests.db_testing import (
        pg_createdb, pg_restore, DB_DUMP_TYPES,
        swh_db_version
    )

    logger.debug('db_init %s dn_name=%s', module, db_name)
    dump_files = []

    for modname in module:
        dump_files.extend(get_sql_for_package(modname))

    if create_db:
        # Create the db (or fail silently if already existing)
        pg_createdb(db_name, check=False)
    # Try to retrieve the db version if any
    db_version = swh_db_version(db_name)
    if not db_version:  # Initialize the db
        dump_files = [(x, DB_DUMP_TYPES[path.splitext(x)[1]])
                      for x in dump_files]
        for dump, dtype in dump_files:
            click.secho('Loading {}'.format(dump), fg='yellow')
            pg_restore(db_name, dump, dtype)

        db_version = swh_db_version(db_name)

    # TODO: Ideally migrate the version from db_version to the latest
    # db version

    click.secho('DONE database is {} version {}'.format(db_name, db_version),
                fg='green', bold=True)


def get_sql_for_package(modname):
    from importlib import import_module
    from swh.core.utils import numfile_sortkey as sortkey

    if not modname.startswith("swh."):
        modname = "swh.{}".format(modname)
    try:
        m = import_module(modname)
    except ImportError:
        raise click.BadParameter("Unable to load module {}".format(modname))

    sqldir = path.join(path.dirname(m.__file__), "sql")
    if not path.isdir(sqldir):
        raise click.BadParameter(
            "Module {} does not provide a db schema "
            "(no sql/ dir)".format(modname))
    return list(sorted(glob.glob(path.join(sqldir, "*.sql")), key=sortkey))



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from zipfile import ZipFile

from swh.core import tarball


def test_is_tarball(tmp_path):

    nozip = tmp_path / 'nozip.zip'
    nozip.write_text('Im no zip')

    assert tarball.is_tarball(str(nozip)) is False

    notar = tmp_path / 'notar.tar'
    notar.write_text('Im no tar')

    assert tarball.is_tarball(str(notar)) is False

    zipfile = tmp_path / 'truezip.zip'
    with ZipFile(str(zipfile), 'w') as myzip:
        myzip.writestr('file1.txt', 'some content')

    assert tarball.is_tarball(str(zipfile)) is True


def test_compress_uncompress_zip(tmp_path):
    tocompress = tmp_path / 'compressme'
    tocompress.mkdir()

    for i in range(10):
        fpath = tocompress / ('file%s.txt' % i)
        fpath.write_text('content of file %s' % i)

    zipfile = tmp_path / 'archive.zip'
    tarball.compress(str(zipfile), 'zip', str(tocompress))

    assert tarball.is_tarball(str(zipfile))

    destdir = tmp_path / 'destdir'
    tarball.uncompress(str(zipfile), str(destdir))

    lsdir = sorted(x.name for x in destdir.iterdir())
    assert ['file%s.txt' % i for i in range(10)] == lsdir


def test_compress_uncompress_tar(tmp_path):
    tocompress = tmp_path / 'compressme'
    tocompress.mkdir()

    for i in range(10):
        fpath = tocompress / ('file%s.txt' % i)
        fpath.write_text('content of file %s' % i)

    tarfile = tmp_path / 'archive.tar'
    tarball.compress(str(tarfile), 'tar', str(tocompress))

    assert tarball.is_tarball(str(tarfile))

    destdir = tmp_path / 'destdir'
    tarball.uncompress(str(tarfile), str(destdir))

    lsdir = sorted(x.name for x in destdir.iterdir())
    assert ['file%s.txt' % i for i in range(10)] == lsdir



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import click
import logging
import pkg_resources

LOG_LEVEL_NAMES = ['NOTSET', 'DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL']

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])

logger = logging.getLogger(__name__)


class AliasedGroup(click.Group):
    'A simple Group that supports command aliases'

    @property
    def aliases(self):
        if not hasattr(self, '_aliases'):
            self._aliases = {}
        return self._aliases

    def get_command(self, ctx, cmd_name):
        return super().get_command(ctx, self.aliases.get(cmd_name, cmd_name))

    def add_alias(self, name, alias):
        if not isinstance(name, str):
            name = name.name
        self.aliases[alias] = name


@click.group(context_settings=CONTEXT_SETTINGS, cls=AliasedGroup)
@click.option('--log-level', '-l', default='INFO',
              type=click.Choice(LOG_LEVEL_NAMES),
              help="Log level (default to INFO)")
@click.pass_context
def swh(ctx, log_level):
    """Command line interface for Software Heritage.
    """
    log_level = logging.getLevelName(log_level)
    logging.root.setLevel(log_level)
    ctx.ensure_object(dict)
    ctx.obj['log_level'] = log_level


def main():
    logging.basicConfig()
    # load plugins that define cli sub commands
    for entry_point in pkg_resources.iter_entry_points('swh.cli.subcommands'):
        cmd = entry_point.load()
        swh.add_command(cmd, name=entry_point.name)

    return swh(auto_envvar_prefix='SWH')


if __name__ == '__main__':
    main()
